# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import print_function

from time import time
py_start_t = time()
"""Approximate start time of the python code"""

import logging
logger = logging.getLogger(__name__)

import os
import threading
import warnings
import inspect

from os.path import isfile, join, dirname

from colorama import Fore

from spyne.util.six import StringIO
from spyne.util.color import R, DARK_R
from spyne.store.relational.util import database_exists, create_database

from sqlalchemy import MetaData

from neurons.daemon.config import FileStore, ServiceDisabled, ServiceDaemon, \
    RelationalStore, LdapStore, Server


def get_package_version(pkg_name):
    try:
        import pkg_resources
        return pkg_resources.get_distribution(pkg_name).version
    except Exception:
        return 'unknown'


def _print_version(config):
    myver = get_package_version(config.name)
    if myver == 'unknown':
        print("Package '%s' version could not be determined. Please make "
              "sure a root package name is passed as daemon_name to the main "
              "function and also the package is correctly installed.")
    else:
        print("This is %s-%s" % (config.name, myver))

    print()
    print("Also:")
    print(" * lxml-%s" % get_package_version('lxml'))
    print(" * pytz-%s" % get_package_version('pytz'))
    print(" * spyne-%s" % get_package_version('spyne'))
    print(" * pyyaml-%s" % get_package_version('pyyaml'))
    print(" * neurons-%s" % get_package_version('neurons'))
    print(" * twisted-%s" % get_package_version('twisted'))
    print(" * msgpack-%s" % get_package_version('msgpack-python'))
    print(" * pycrypto-%s" % get_package_version('pycrypto'))
    print(" * werkzeug-%s" % get_package_version('werkzeug'))
    print(" * txpostgres-%s" % get_package_version('txpostgres'))
    print(" * SQLAlchemy-%s" % get_package_version('SQLAlchemy'))
    print()

    return 0  # to force exit


def _write_wsdl(config):
    from lxml import etree

    from spyne.interface.wsdl import Wsdl11
    from spyne.test.sort_wsdl import sort_wsdl
    from spyne.util.appreg import applications

    for (tns, name), appdata in applications.items():
        appdata.app.transport = "no_transport_at_all"
        wsdl = Wsdl11(appdata.app.interface)
        wsdl.build_interface_document('hxxp://invalid_url')
        doc = wsdl.get_interface_document()
        elt = etree.parse(StringIO(doc))
        sort_wsdl(elt)

        file_name = join(config.write_wsdl, 'wsdl.%s.xml' % name)

        try:
            os.makedirs(dirname(file_name))
        except OSError:
            pass

        try:
            with open(file_name, 'w') as f:
                f.write(etree.tostring(elt, pretty_print=True))
        except Exception as e:
            print("Error:", e)
            return -1

        print(file_name, "written.")

    return 0  # to force exit


def _write_xsd(config):
    from lxml import etree

    from spyne.interface.xml_schema import XmlSchema
    from spyne.util.appreg import applications

    for (tns, name), appdata in applications.items():
        xsd = XmlSchema(appdata.app.interface)
        xsd.build_interface_document()
        schemas = xsd.get_interface_document()

        dir_name = join(config.write_xsd, 'schema.%s' % name)

        try:
            os.makedirs(dir_name)
        except OSError:
            pass

        for k, v in schemas.items():
            file_name = os.path.join(dir_name, "%s.xsd" % k)

            try:
                with open(file_name, 'wb') as f:
                    etree.ElementTree(v).write(f, pretty_print=True)

            except Exception as e:
                print("Error:", e)
                return -1

            print("written",file_name, "for ns", appdata.app.interface.nsmap[k])

    return 0  # to force exit


def _do_bootstrap(config, init, bootstrap, bootstrapper):
    if bootstrap is None:
        bootstrap = bootstrapper(init)
    else:
        config.apply()

    assert callable(bootstrap), \
                      "'bootstrap' must be a callable. It's %r." % bootstrap

    # perform bootstrap
    retval = bootstrap(config)

    if retval is None:
        return 0  # to force exit
    return retval


def _do_drop_all_tables(config, init):
    config.log_queries = True
    config.apply()

    import neurons
    meta = neurons.TableModel.Attributes.sqla_metadata
    meta.reflect()

    meta.drop_all()

    return 0


def _do_start_shell(config):
    # Import db handle, session and other useful stuff to the shell's scope
    db = None
    if isinstance(config, ServiceDaemon):
        db = config.get_main_store()

    # so that there is a db session handy in the shell
    session = db.Session()

    # these are just useful to have in a dev. shell
    import IPython

    header = (
        "Database handle is:  db\n"
        "There's also an open session: session\n"
        "Imported packages: traceback, inspect, sys\n"
        "Imported functions: pprint(), pformat()"
    )

    # start the kind of shell requested by user
    if config.shell:
        return IPython.embed(header=header)

    if config.ikernel:
        return IPython.embed_kernel()


def _set_real_factory(lp, subconfig, factory):
    # lp = listening port -- what endpoint.listen()'s return value ends up as
    if hasattr(lp.factory, 'wrappedFactory'):
        import twisted.protocols.tls
        assert isinstance(lp.factory, twisted.protocols.tls.TLSMemoryBIOFactory)
        target_factory = lp.factory.wrappedFactory

    else:
        target_factory = lp.factory

    subconfig.color = Fore.GREEN

    assert isinstance(target_factory, Server.FactoryProxy)
    target_factory.real_factory = factory

    logger.info("%s Service ready with factory %r",
                                                subconfig.colored_name, factory)


def _inner_main(config, init, bootstrap, bootstrapper):
    # if requested, print version and exit
    if config.version:
        return _print_version(config)

    # if requested, perform bootstrap and exit
    if config.bootstrap:
        return _do_bootstrap(config, init, bootstrap, bootstrapper)

    if isinstance(config, ServiceDaemon):
        # if requested, drop all tables and exit
        if config.drop_all_tables:
            return _do_drop_all_tables(config, init)

    config.apply()
    logger.info("%s config valid, initializing services...", config.name)

    # initialize main table model
    if isinstance(config, ServiceDaemon):
        if config.main_store is None:
            config.main_store = 'sql_main'

        if config.debug_reactor:
            config.add_reactor_checks()

        from neurons import TableModel
        TableModel.Attributes.sqla_metadata.bind = \
                                                  config.get_main_store().engine

    # initialize applications
    items = init(config)
    if hasattr(items, 'items'):  # if it's a dict
        items = items.items()

    # apply app-specific config
    for k, v in items:
        disabled = False
        if k in config.services:
            disabled = config.services[k].disabled

        if disabled:
            logger.info("%s Service disabled.", DARK_R('[%s]' % (k,)))
            continue

        try:
            if v.force is not None:
                if k in config.services:
                    oldconfig = config.services[k]
                    subconfig = config.services[k] = v.force
                    if oldconfig.d is not None:
                        subconfig.d = oldconfig.d

                    if oldconfig.listener is not None:
                        subconfig.listener = oldconfig.listener

                else:
                    subconfig = config.services[k] = v.force

                logger.info("%s Configuration initialized from "
                                   "hard-coded object.", subconfig.colored_name)

            else:
                k_was_there = k in config.services
                subconfig = config.services.getwrite(k, v.default)

                if k_was_there:
                    logger.info("%s Configuration initialized from file.",
                                                         subconfig.colored_name)
                else:
                    logger.info("%s Configuration initialized from default.",
                                                         subconfig.colored_name)

            factory = v.init(config)

        except ServiceDisabled:
            logger.info("%s Service disabled.", R('[%s]' % (k,)))
            continue

        if not isinstance(subconfig, Server):
            continue

        if subconfig.d is not None:
            if subconfig.listener is None:
                subconfig.d.addCallback(_set_real_factory, subconfig, factory)

            else:
                _set_real_factory(subconfig.listener, subconfig, factory)

        else:
            subconfig.listen() \
                .addCallback(_set_real_factory, subconfig, factory)

    # if requested, write interface documents and exit
    if isinstance(config, ServiceDaemon):
        if config.write_wsdl:
            return _write_wsdl(config)

        if config.write_xsd:
            return _write_xsd(config)

    if config.write_config:
        config.do_write_config()
        return 0

    # Perform schema migrations
    from neurons.version import Version
    Version.migrate_all(config)

    # if requested, drop to shell
    if config.shell or config.ikernel:
        ret = _do_start_shell(config)
        if ret is None:
            return 0


class Bootstrapper(object):
    """Creates all databases"""

    def __init__(self, init):
        self.init = init
        self.meta_reflect = MetaData()

    def before_tables(self, config):
        pass

    def after_tables(self, config):
        pass

    def create_relational(self, store):
        if database_exists(store.conn_str):
            print(store.conn_str, "already exists.")
            return

        create_database(store.conn_str)
        print(store.conn_str, "did not exist, created.")

    def __call__(self, config):
        # we are printing stuff here in case the log goes to a log file and the
        # poor ops guy can't see a thing
        for store in config.stores.values():
            if isinstance(store, RelationalStore):
                self.create_relational(store)

            elif isinstance(store, LdapStore):
                warnings.warn("LDAP bootstrap is not implemented.")

            elif isinstance(store, FileStore):
                try:
                    os.makedirs(store.path)
                    print("File store", store.name, "directory", store.path,
                                                            'has been created.')
                except OSError:
                    print("File store", store.name, "directory", store.path,
                                                              'already exists.')

            else:
                raise ValueError(store)

        config.apply(daemonize=False)

        main_engine = config.get_main_store().engine

        # reflect database just in case -- can be useful while bootstrapping
        self.meta_reflect.reflect(bind=main_engine)
        print("Reflection")

        # Run init so that all relevant models get imported
        self.init(config)
        print("Init")

        from neurons.model import TableModel
        TableModel.Attributes.sqla_metadata.bind = main_engine

        # Init schema versions
        from neurons.version import Version
        Version.migrate_all(config)

        self.before_tables(config)

        TableModel.Attributes.sqla_metadata.create_all(checkfirst=True)
        print("All tables created.")

        self.after_tables(config)

        from spyne.util.color import G
        print(G("Bootstrap complete."))


def _set_reactor_thread():
    import neurons
    neurons.REACTOR_THREAD = threading.current_thread()
    neurons.REACTOR_THREAD_ID = neurons.REACTOR_THREAD.ident
    neurons.is_reactor_thread = neurons._base._is_reactor_thread


def _compile_mappers():
    logger.info("Compiling object mappers...")
    from sqlalchemy.orm import compile_mappers
    compile_mappers()


def boot(config_name, argv, init, bootstrap=None,
                bootstrapper=Bootstrapper, cls=ServiceDaemon, daemon_name=None):
    """Boots the daemon. The signature is the same as the ``main()`` function in
    this module.

    :param config_name: Configuration file name. .yaml suffix is appended.
    :param argv: A sequence of command line arguments.
    :param init: A callable that returns the init dict.
    :param bootstrap: A callable that bootstraps daemon's environment.
        It's DEPRECATED in favor of bootstrapper.
    :param bootstrapper: A factory for a callable that bootstraps daemon's
        environment. This is supposed to be run once for every new deployment.
    :param cls: a class:`Daemon` subclass
    :param daemon_name: Daemon name. If ``None``, ``config_name`` is used.
    :return: The return code to be passed to sys.exit() and the daemon config
        object. If the return code is None, it's OK to proceed with running the
        daemon.
    """

    config = cls.parse_config(config_name, argv)
    if config.help:
        from neurons.daemon.cli import spyne_to_argparse

        print(spyne_to_argparse(cls, ignore_defaults=False).format_help())
        return 0, config

    if config.name is None:
        config.name = config_name

    if daemon_name is not None:
        config.name = daemon_name

    # FIXME: Any better ideas?
    has_services = hasattr(config, '_services')
    has_stores = hasattr(config, '_stores')

    services = None
    if has_services:
        services = list(config._services)

    stores = None
    if has_stores:
        stores = list(config._stores)

    try:
        retcode = _inner_main(config, init, bootstrap, bootstrapper)

        # if _inner_main did something other than initializing daemons
        if retcode is not None:
            return retcode, config

    finally:
        if not isfile(config.config_file):
            config.do_write_config()
            logger.info("Writing configuration to: '%s'", config.config_file)

        elif has_services and services != config._services:
            config.do_write_config()
            logger.info("Updating configuration file because new services were "
                                                                     "detected")

        elif has_stores and stores != config._stores:
            config.do_write_config()
            logger.info("Updating configuration file because new stores were "
                                                                     "detected")

        # FIXME: could someone need these during bootstrap above?
        if config.uuid is None:
            config.uuid = config.gen_uuid()
            config.do_write_config()
            logger.info("Updating configuration file because new uuid was "
                                                                    "generated")

        if config.secret is None:
            config.secret = config.gen_secret()
            config.do_write_config()
            logger.info("Updating configuration file because new secret was "
                                                                    "generated")

    return None, config

def _log_ready(config_name, orig_stack, py_start_t, func_start_t):
    import resource

    try:
        import psutil
        proc_start_t = psutil.Process(os.getpid()).create_time()
    except ImportError:
        proc_start_t = '?'

    max_rss = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1000.0

    package_name = config_name

    frame, file_name, line_num, func_name, lines, line_id = orig_stack[1]
    module = inspect.getmodule(frame)
    if module is not None:
        package_name = module.__name__.split('.')[0]

    logger.info(
        "%s version %s ready. Max RSS: %.1fmb uptime: %s import: %.2fs "
                                                              "main: %.2fs",
        config_name, get_package_version(package_name),
        max_rss,
        '[?psutil?]' if proc_start_t == '?'
                                    else '%.2fs' % (time() - proc_start_t,),
        time() - py_start_t,
        time() - func_start_t,
    )


def main(config_name, argv, init, bootstrap=None,
                bootstrapper=Bootstrapper, cls=ServiceDaemon, daemon_name=None):
    """Boots and runs the daemon, making it ready to accept requests. This is a
    typical main function for daemons. If you just want to boot the daemon
    and take care of running it yourself, see the ``boot()`` function.

    :param config_name: Configuration file name. .yaml suffix is appended.
    :param argv: A sequence of command line arguments.
    :param init: A callable that returns the init dict.
    :param bootstrap: A callable that bootstraps daemon's environment.
        It's deprecated in favor of bootstrapper.
    :param bootstrapper: A factory for a callable that bootstraps daemon's
        environment. This is supposed to be run once for every new deployment.
    :param cls: a class:`Daemon` subclass
    :param daemon_name: Daemon name. If ``None``, ``config_name`` is used.
    :return: Exit code of the daemon as int.
    """

    func_start_t = time()
    """Start time of post-import initialization code"""

    retcode, config = boot(config_name, argv, init, bootstrap,
                                                 bootstrapper, cls, daemon_name)

    # at this point it's safe to import the reactor (or anything else from
    # twisted) because the decision on whether to fork has already been made.
    from twisted.internet import reactor
    from twisted.internet.task import deferLater

    deferLater(reactor, 0, _compile_mappers) \
        .addErrback(lambda err: logger.error("%s", err.getTraceback()))

    deferLater(reactor, 0, _set_reactor_thread) \
        .addErrback(lambda err: logger.error("%s", err.getTraceback()))

    deferLater(reactor, 0, _log_ready,
                       config_name, inspect.stack(), py_start_t, func_start_t) \
        .addErrback(lambda err: logger.error("%s", err.getTraceback()))

    # this needs to be done as late as possible to capture the highest number of
    # watched (ie imported) modules.
    if config.autoreload:
        from spyne.util.autorel import AutoReloader
        frequency = 0.5
        autorel = AutoReloader(frequency=frequency)
        assert autorel.start() is not None
        num_files = len(autorel.sysfiles() | autorel.files)
        logger.info("Auto reloader init success: Watching %d files "
                                      "every %g seconds.", num_files, frequency)

    if config.dry_run:
        return 0

    return reactor.run()



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# retrieved from: http://code.activestate.com/recipes/278731/
# Licensed under the PSF License

"""Disk And Execution MONitor (Daemon)

Configurable daemon behaviors:

   1.) The current working directory set to the "/" directory.
   2.) The current file creation mode mask set to 0.
   3.) Close all open files (1024). 
   4.) Redirect standard I/O streams to "/dev/null".

A failed call to fork() now raises an exception.

References:
   1) Advanced Programming in the Unix Environment: W. Richard Stevens
   2) Unix Programming Frequently Asked Questions:
         http://www.erlenstar.demon.co.uk/unix/faq_toc.html
"""

__author__ = "Chad J. Schroeder"
__copyright__ = "Copyright (C) 2005 Chad J. Schroeder"

__revision__ = "$Id$"
__version__ = "0.2"


import os
import logging
logger = logging.getLogger(__name__)


# Default maximum for the number of available file descriptors.
MAXFD = 1024


def daemonize_do(umask=0, workdir='/', maxfd=None, redirect_to=os.devnull):
    """Detach a process from the controlling terminal and run it in the
    background as a daemon.

    :param umask: File mode creation mask of the daemon.
    :param workdir: Default working directory for the daemon.
    :param maxfd: Default maximum for the number of available file descriptors.
    :param redirect_to: The target for standard I/O file descriptors. By
        default, it's `/dev/null`.
    """

    # Fork a child process so the parent can exit.  This returns control to
    # the command-line or shell.  It also guarantees that the child will not
    # be a process group leader, since the child receives a new process ID
    # and inherits the parent's process group ID.  This step is required
    # to insure that the next call to os.setsid is successful.
    pid = os.fork()

    if pid == 0:      # The first child.
        # To become the session leader of this new session and the process group
        # leader of the new process group, we call os.setsid().  The process is
        # also guaranteed not to have a controlling terminal.
        os.setsid()

        # Is ignoring SIGHUP necessary?
        #
        # It's often suggested that the SIGHUP signal should be ignored before
        # the second fork to avoid premature termination of the process.  The
        # reason is that when the first child terminates, all processes, e.g.
        # the second child, in the orphaned group will be sent a SIGHUP.
        #
        # "However, as part of the session management system, there are exactly
        # two cases where SIGHUP is sent on the death of a process:
        #
        # 1) When the process that dies is the session leader of a session that
        #      is attached to a terminal device, SIGHUP is sent to all processes
        #      in the foreground process group of that terminal device.
        # 2) When the death of a process causes a process group to become
        #      orphaned, and one or more processes in the orphaned group are
        #      stopped, then SIGHUP and SIGCONT are sent to all members of the
        #      orphaned group." [2]
        #
        # The first case can be ignored since the child is guaranteed not to
        # have a controlling terminal. The second case isn't so easy to
        # dismiss.
        # The process group is orphaned when the first child terminates and
        # POSIX.1 requires that every STOPPED process in an orphaned process
        # group be sent a SIGHUP signal followed by a SIGCONT signal.  Since the
        # second child is not STOPPED though, we can safely forego ignoring the
        # SIGHUP signal. In any case, there are no ill-effects if it is ignored.
        #
        import signal           # Set handlers for asynchronous events.
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

        # Fork a second child and exit immediately to prevent zombies.  This
        # causes the second child process to be orphaned, making the init
        # process responsible for its cleanup. And, since the first child is
        # a session leader without a controlling terminal, it's possible for
        # it to acquire one by opening a terminal in the future (System V-
        # based systems).  This second fork guarantees that the child is no
        # longer a session leader, preventing the daemon from ever acquiring
        # a controlling terminal.
        pid = os.fork()    # Fork a second child.

        if pid == 0:    # The second child.
            # Since the current working directory may be a mounted filesystem,
            # we avoid the issue of not being able to unmount the filesystem at
            # shutdown time by changing it to the root directory.
            logger.debug("Change working directory to: %s", workdir)
            os.chdir(workdir)

            # We probably don't want the file mode creation mask inherited from
            # the parent, so we give the child complete control over
            # permissions.
            os.umask(umask)

        else:
            # exit() or _exit()?  See below.
            os._exit(0)    # Exit parent (the first child) of the second child.
    else:
        # exit() or _exit()?
        # _exit is like exit(), but it doesn't call any functions registered
        # with atexit (and on_exit) or any registered signal handlers.  It also
        # closes any open file descriptors.  Using exit() may cause all stdio
        # streams to be flushed twice and any temporary files may be
        # unexpectedly removed.  It's therefore recommended that child branches
        # of a fork() and the parent branch(es) of a daemon use _exit().
        os._exit(0)    # Exit parent of the first child.

    # Close all open file descriptors.  This prevents the child from keeping
    # open any file descriptors inherited from the parent.  There is a variety
    # of methods to accomplish this task.  Three are listed below.
    #
    # Try the system configuration variable, SC_OPEN_MAX, to obtain the maximum
    # number of open file descriptors to close.  If it doesn't exists, use
    # the default value (configurable).
    #
    # try:
    #    maxfd = os.sysconf("SC_OPEN_MAX")
    # except (AttributeError, ValueError):
    #    maxfd = MAXFD
    #
    # OR
    #
    # if (os.sysconf_names.has_key("SC_OPEN_MAX")):
    #    maxfd = os.sysconf("SC_OPEN_MAX")
    # else:
    #    maxfd = MAXFD
    #
    # OR
    #
    # Use the getrlimit method to retrieve the maximum file descriptor number
    # that can be opened by this process.  If there is not limit on the
    # resource, use the default value.
    #
    import resource  # Resource usage information.

    if maxfd is None:
        soft_fd_limit, hard_fd_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
        if hard_fd_limit != resource.RLIM_INFINITY:
            maxfd = hard_fd_limit
        else:
            maxfd = MAXFD

    # Iterate through and close all file descriptors.
    for fd in range(0, maxfd):
        try:
            os.close(fd)
        except OSError:    # ERROR, fd wasn't open to begin with (ignored)
            pass

    # Redirect the standard I/O file descriptors to the specified file.  Since
    # the daemon has no controlling terminal, most daemons redirect stdin,
    # stdout, and stderr to /dev/null.  This is done to prevent side-effects
    # from reads and writes to the standard I/O file descriptors.

    # This call to open is guaranteed to return the lowest file descriptor,
    # which will be 0 (stdin), since it was closed above.
    os.open(redirect_to, os.O_RDWR)    # standard input (0)

    # Duplicate standard input to standard output and standard error.
    os.dup2(0, 1)            # standard output (1)
    os.dup2(0, 2)            # standard error (2)

    return 0



#!/usr/bin/env python
# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Burak Arslan <burak.arslan@arskom.com.tr>,
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import absolute_import, print_function

import logging
logging.basicConfig(level=logging.DEBUG)

import re
import unittest

from decimal import Decimal as D
from datetime import date, time, datetime

from lxml import etree, html

from spyne import Application, NullServer, Unicode, Service, rpc, Decimal, \
    Boolean, Date, Time, DateTime, Integer, ComplexModel, Array, Double, \
    Mandatory as M, AnyHtml, AnyXml
from spyne.util.test import show
from spyne.protocol.html import HtmlMicroFormat

from neurons.form import HtmlForm, PasswordWidget, Tab, HrefWidget, \
    ComboBoxWidget, ComplexHrefWidget, ParentHrefWidget
from neurons.form.test import strip_ns
from neurons.form.const import T_TEST
from neurons.form.form import Fieldset


def _test_type(cls, inst):
    # silence bogus warnings
    from spyne.util import appreg; appreg.applications.clear()

    class SomeService(Service):
        @rpc(_returns=cls, _body_style='bare')
        def some_call(ctx):
            return inst

    prot = HtmlForm(cloth=T_TEST)
    app = Application([SomeService], 'some_ns', out_protocol=prot)

    null = NullServer(app, ostr=True)

    ret = ''.join(null.service.some_call())
    try:
        elt = html.fromstring(ret)
    except:
        print(ret)
        raise

    show(elt, stdout=False)
    elt = elt.xpath('//form')[0]  # get the form tag inside the body tag
    elt = strip_ns(elt)  # get rid of namespaces to simplify xpaths in tests

    print(etree.tostring(elt, pretty_print=True))

    return elt


def _test_type_no_root_cloth(cls, inst):
    from spyne.util import appreg; appreg.applications.clear()

    class SomeService(Service):
        @rpc(_returns=cls, _body_style='bare')
        def some_call(ctx):
            return inst

    prot = HtmlForm()
    app = Application([SomeService], 'some_ns', out_protocol=prot)

    null = NullServer(app, ostr=True)
    elt = etree.fromstring(''.join(null.service.some_call()))
    show(elt)

    return elt


class TestFormPrimitive(unittest.TestCase):
    def test_unicode(self):
        v = 'foo'
        elt = _test_type(Unicode(100, type_name="tn"), v).xpath('div/input')[0]
        assert elt.attrib['type'] == 'text'
        assert elt.attrib['value'] == v

    def test_unicode_boundless(self):
        v = 'foo'
        elt = _test_type(Unicode, v).xpath('div/textarea')[0]
        assert elt.tag == 'textarea'
        assert elt.text == v

    def test_unicode_null(self):
        v = None
        elt = _test_type(Unicode(100, type_name="tn"), v).xpath('div/input')[0]
        assert elt.attrib['type'] == 'text'
        assert not ('value' in elt.attrib)

    def test_unicode_values(self):
        v = 'a'
        cls = Unicode(100, type_name="tn", values=list('abcd'))
        assert not cls.get_type_name() is Unicode.Empty
        elt = _test_type(cls, v).xpath('div/select')[0]
        assert elt.tag == 'select'
        assert elt.xpath("option/@value") == [''] + list('abcd')
        assert elt.xpath("option[@selected]/text()") == [v]

    def test_integer_values(self):
        v = 3
        cls = Integer(type_name="tn", values=list(range(5)))
        assert not cls.get_type_name() is Unicode.Empty
        elt = _test_type(cls, v).xpath('div/select')[0]
        assert elt.tag == 'select'
        assert elt.xpath("option/@value") == ['']+[str(vvv) for vvv in range(5)]
        assert elt.xpath("option[@selected]/text()") == [str(v)]

    def test_integer_values_mandatory_field(self):
        v = 3
        cls = M(Integer(type_name="tn", values=list(range(5))))
        assert not cls.get_type_name() is Unicode.Empty
        elt = _test_type(cls, v).xpath('div/select')[0]
        assert elt.tag == 'select'
        assert elt.xpath("option/@value") == [str(vvv) for vvv in range(5)]
        assert elt.xpath("option[@selected]/text()") == [str(v)]

    def test_unicode_password(self):
        elt = _test_type(Unicode(64, prot=PasswordWidget()), None)
        elt = elt.xpath('div/input')[0]
        assert elt.attrib['type'] == 'password'

    def test_decimal(self):
        elt = _test_type(Decimal, D('0.1')).xpath('div/input')[0]
        assert elt.attrib['type'] == 'number'
        assert elt.attrib['step'] == 'any'

    # FIXME: enable this after fixing the relevant Spyne bug
    def _test_decimal_step(self):
        elt = _test_type(Decimal(fraction_digits=4), D('0.1'))
        elt = elt.xpath('div/input')[0]

        assert elt.attrib['step'] == '0.0001'

    def test_boolean_true(self):
        elt = _test_type(Boolean, True).xpath('div/input')[0]
        assert 'checked' in elt.attrib

    def test_boolean_false(self):
        elt = _test_type(Boolean, False).xpath('div/input')[0]
        assert not ('checked' in elt.attrib)

    def test_date(self):
        elt = _test_type(Date, date(2013, 12, 11)).xpath('div/input')[0]
        assert elt.attrib['value'] == '2013-12-11'
        # FIXME: Need to find a way to test the generated js

    def test_time(self):
        elt = _test_type(Time, time(10, 9, 8)).xpath('div/input')[0]
        assert elt.attrib['value'] == '10:09:08'
        # FIXME: Need to find a way to test the generated js

    def test_datetime(self):
        v = datetime(2013, 12, 11, 10, 9, 8)
        script = _test_type(DateTime, v).xpath('div/script/text()')[0]
        assert v.isoformat() in script
        # FIXME: Need to find a better way to test the generated js

    def test_array_datetime(self):
        v = datetime(2013, 12, 11, 10, 9, 8)
        _ = _test_type(Array(DateTime), [v]).xpath('div/script/text()')[0]
        # FIXME: test something!

    def test_datetime_format_split(self):
        ret = HtmlForm._split_datetime_format('%Y-%m-%d %H:%M:%S')
        assert ret == ('yy-mm-dd', 'HH:mm:ss')

        ret = HtmlForm._split_datetime_format('%Y-%m-%d %H:%M')
        assert ret == ('yy-mm-dd', 'HH:mm')

        ret = HtmlForm._split_datetime_format('%Y-%m-%d')
        assert ret == ('yy-mm-dd', '')

        ret = HtmlForm._split_datetime_format('%H:%M:%S')
        assert ret == ('', 'HH:mm:ss')

    def test_integer(self):
        v = 42
        elt = _test_type(Integer, v).xpath('div/input')[0]
        assert elt.attrib['value'] == str(v)

    def test_integer_constraints(self):
        elt = _test_type(Integer(gt=0), None).xpath('div/input')[0]
        assert elt.attrib['min'] == '1'

        elt = _test_type(Integer(ge=0), None).xpath('div/input')[0]
        assert elt.attrib['min'] == '0'

    def test_decimal_constraints(self):
        elt = _test_type(Decimal(ge=0), None).xpath('div/input')[0]
        assert elt.attrib['min'] == '0'

        elt = _test_type(Decimal(gt=0), None).xpath('div/input')[0]
        assert elt.attrib['min'] == '1E-308'

    def test_integer_none(self):
        v = None
        elt = _test_type(Integer, v).xpath('div/input')[0]
        assert not 'value' in elt.attrib

    def test_hidden_input(self):
        v = 5
        elt = _test_type(Integer(hidden=True), v).xpath('input')[0]

        assert elt.attrib['type'] == "hidden"
        assert elt.attrib['value'] == str(v)

    def test_hidden_input_pa(self):
        v = "punk"
        elt = _test_type(Unicode(pa={HtmlForm:dict(hidden=True)}), v) \
                                                              .xpath('input')[0]
        assert elt.attrib['type'] == "hidden"
        assert elt.attrib['value'] == v

    def test_hidden_input_none(self):
        v = None
        elt = _test_type(Integer(hidden=True), v).xpath('input')[0]

        assert elt.attrib['type'] == "hidden"
        assert not 'value' in elt.attrib

    def test_html(self):
        v = "<html><head></head><body><p>test</p></body></html>"
        elt = _test_type(AnyHtml, v).xpath('div/textarea')[0]

        assert elt.text == str(v)

    def test_xml(self):
        v = "<html><head/><body><p>test</p></body></html>"
        elt = _test_type(AnyXml, v).xpath('div/textarea')[0]

        assert elt.text == str(v)


class TestFormComplex(unittest.TestCase):
    # all complex objects serialize to forms with fieldsets. that's why we
    # always run xpaths on elt[0] i.e. inside the fieldset where the data we're
    # after is, instead of running longer xpath queries.

    def test_simple(self):
        class SomeObject(ComplexModel):
            _type_info = [
                ('i', Integer),
                ('s', Unicode(64)),
            ]

        v = SomeObject(i=42, s="Arthur")
        elt = _test_type(SomeObject, v)[0]
        assert elt.xpath('div/input/@value') == ['42', 'Arthur']
        assert elt.xpath('div/input/@name') == ['i', 's']

    def test_no_fieldset(self):
        class SomeObject(ComplexModel):
            _type_info = [
                ('i', Integer),
                ('s', Unicode(64)),
            ]
        SomeObject = SomeObject.customize(no_fieldset=True)
        v = SomeObject(i=42, s="Arthur")
        elt = _test_type(SomeObject, v)
        assert elt.xpath('div/input/@value') == ['42', 'Arthur']
        assert elt.xpath('div/input/@name') == ['i', 's']

    def test_nested(self):
        class InnerObject(ComplexModel):
            _type_info = [
                ('s', Unicode(64)),
            ]
        class OuterObject(ComplexModel):
            _type_info = [
                ('i', InnerObject),
                ('d', Double),
            ]

        v = OuterObject(i=InnerObject(s="Arthur"), d=3.1415)
        elt = _test_type(OuterObject, v)[0]

        # it's a bit risky doing this with doubles
        assert elt.xpath('div/input/@value') == ['3.1415']
        assert elt.xpath('div/input/@name') == ['d']
        assert elt.xpath('fieldset/div/input/@value') == ['Arthur']
        assert elt.xpath('fieldset/div/input/@name') == ['i.s']

    def test_fieldset(self):
        fset_one = Fieldset("One")
        fset_two = Fieldset("Two")
        class SomeObject(ComplexModel):
            _type_info = [
                ('i0', Integer),
                ('s0', Unicode(64)),
                ('i1', Integer(fieldset=fset_one)),
                ('s1', Unicode(64, fieldset=fset_one)),
                ('i2', Integer(fieldset=fset_two)),
                ('s2', Unicode(64, fieldset=fset_two)),
            ]

        v = SomeObject(
            i0=42, s0="Arthur",
            i1=42, s1="Arthur",
            i2=42, s2="Arthur",
        )
        elt = _test_type(SomeObject, v)[0]
        assert elt.xpath('div/input/@value') == ['42', 'Arthur']
        assert elt.xpath('div/input/@name') == ['i0', 's0']
        assert elt.xpath('fieldset/div/input/@value') == ['42', 'Arthur',
                                                          '42', 'Arthur']
        assert elt.xpath('fieldset/div/input/@name') == ['i1', 's1', 'i2', 's2']

    def test_tab(self):
        tab1 = Tab("One")
        tab2 = Tab("Two")
        class SomeObject(ComplexModel):
            _type_info = [
                ('i0', Integer),
                ('i1', Integer(tab=tab1)),
                ('i2', Integer(tab=tab2)),
            ]

        v = SomeObject(i0=14, i1=28, i2=56)
        elt = _test_type(SomeObject, v)[0]
        assert elt.xpath('div/input/@value') == ['14']
        assert elt.xpath('div/input/@name') == ['i0']

        assert elt.xpath('div/ul/li/a/text()') == [tab1.legend, tab2.legend]
        assert elt.xpath('div/ul/li/a/@href') == \
                                          ["#" + tab1.htmlid, "#" + tab2.htmlid]
        assert elt.xpath('div/div/@id') == [tab1.htmlid, tab2.htmlid]
        assert elt.xpath('div/div[@id]/div/input/@name') == ['i1', 'i2']
        assert elt.xpath('div/div[@id]/div/input/@value') == ['28', '56']

        # FIXME: properly test script tags
        assert elt.xpath('div/@id')[0] in elt.xpath('script/text()')[0]

    def test_simple_array(self):
        class SomeObject(ComplexModel):
            _type_info = [
                ('ints', Array(Integer)),
            ]

        v = SomeObject(ints=range(5))
        elt = _test_type(SomeObject, v)[0]

        assert elt.xpath('div/div/div/input/@value') == \
                                                       ['0', '1', '2', '3', '4']
        assert elt.xpath('div/div/button/text()') == ['+', '-'] * 5

        names = elt.xpath('div/div/div/input/@name')
        assert len(names) > 0

        for i, name in enumerate(names):
            assert re.match(r'ints\[0*%d\]' % i, name)


class TestSimpleHrefWidget(object):
    def test_simple(self):
        v = Integer(prot=HrefWidget(href="/some_href?id={}"))
        elt = _test_type(v, 5)
        assert elt.xpath('div/span/a/text()') == ['5']
        assert elt.xpath('div/span/a/@href') == ['/some_href?id=5']


class TestComplexHrefWidget(object):
    def test_simple(self):
        class SomeObject(ComplexModel):
            class Attributes(ComplexModel.Attributes):
                prot = ComplexHrefWidget('s', 'i')

            _type_info = [
                ('i', Integer),
                ('s', Unicode),
            ]

        v = SomeObject(i=42, s="Arthur")
        elt = _test_type(SomeObject, v)

        assert elt.xpath('div/a/text()') == ['Arthur']
        assert elt.xpath('div/a/@href') == ['some_object?i=42']

    def test_granchild(self):
        class SomeOtherObject(ComplexModel):
            _type_info = [
                ('d', Decimal),
            ]

        class SomeObject(ComplexModel):
            class Attributes(ComplexModel.Attributes):
                prot = ComplexHrefWidget('c.d', 'i')

            _type_info = [
                ('i', Integer),
                ('c', SomeOtherObject),
            ]

        v = SomeObject(i=42, c=SomeOtherObject(d=3.14))
        elt = _test_type(SomeObject, v)

        assert elt.xpath('div/a/text()') == ['3.14']
        assert elt.xpath('div/a/@href') == ['some_object?i=42']


class TestParentHrefWidget(object):
    def test_simple(self):
        class SomeObject(ComplexModel):
            class Attributes(ComplexModel.Attributes):
                prot = HtmlMicroFormat()

            _type_info = [
                ('i', Integer),
                ('s', Unicode(prot=ParentHrefWidget('some_object?i={0.i}',
                    anchor_class="some_class"))),
            ]

        v = SomeObject(i=42, s="Arthur")
        elt = _test_type(SomeObject, v)

        assert elt.xpath('.//a[@class="some_class"]/text()') == ['Arthur']
        assert elt.xpath('.//a[@class="some_class"]/@href') == \
                                                            ['some_object?i=42']


class TestComboBoxWidget(object):
    def test_simple_label(self):
        class SomeObject(ComplexModel):
            class Attributes(ComplexModel.Attributes):
                prot = ComboBoxWidget('s', 'i')

            _type_info = [
                ('i', Integer),
                ('s', Unicode),
            ]

        v = SomeObject(i=42, s="Arthur")
        elt = _test_type(SomeObject, v)

        assert elt.xpath('div/label/text()') == ['SomeObject']
        assert elt.xpath('div/select/option/text()') == ['Arthur']
        assert elt.xpath('div/select/option/@value') == ['42']

    def test_simple_nolabel(self):
        class SomeObject(ComplexModel):
            class Attributes(ComplexModel.Attributes):
                prot = ComboBoxWidget('s', 'i', label=False)

            _type_info = [
                ('i', Integer),
                ('s', Unicode),
            ]

        v = SomeObject(i=42, s="Arthur")
        elt = _test_type(SomeObject, v)

        assert elt.xpath('select/option/text()') == ['Arthur']
        assert elt.xpath('select/option/@value') == ['42']


if __name__ == '__main__':
    unittest.main()



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import logging
logger = logging.getLogger(__name__)

from time import time
from inspect import isclass

from twisted.internet.defer import Deferred


def _get_out_object(ctx):
    """Returns out object from context

    :param om: out_message
    :return: out_object
    """
    om = ctx.descriptor.out_message
    oo = None
    if ctx.descriptor.is_out_bare():
        oo = next(iter(ctx.out_object))

    else:  # om is a ComplexModelBase then.
        if len(om._type_info) == 0:
            oo, = ctx.out_object

        elif len(om._type_info) == 1:
            om, = om._type_info.values()
            oo, = ctx.out_object

        else:
            oo = om.get_serialization_instance(ctx.out_object)

    return oo


def t_log_method_return(LogEntry):
    def _on_method_return(ctx):
        if ctx.service_class is not None and ctx.service_class.is_auxiliary():
            return
        ctx.event.call_end = time()

        log_entry = LogEntry()

        omc = None
        if ctx.descriptor.is_out_bare():
            omc = ctx.descriptor.out_message

        elif len(ctx.descriptor.out_message._type_info) == 1:
            omc, = ctx.descriptor.out_message._type_info.values()

        if isclass(omc) and issubclass(omc, Integer) and \
                               isinstance(ctx.out_object[0], six.integer_types):
            log_entry.data_out_int = next(iter(ctx.out_object))

        if ctx.udc.no_persistent_log:
            logger.debug("not logging %r", ctx.descriptor.key)
            ctx.udc.log_entry = None

        else:
            oo = _get_out_object(ctx)
            ctx.udc.log_entry = log_entry

            def _log(ret, ctx, log_entry):
                from twisted.internet.threads import deferToThread
                return deferToThread(_fill_log, ctx, log_entry) \
                    .addCallbacks(lambda _: ret, lambda err: logging.exception(err.value))


            if isinstance(oo, Deferred):
                oo.addCallback(_log, ctx, log_entry)
                oo.addErrback(_log, ctx, log_entry)

            else:
                _log(None, ctx, log_entry)

    return _on_method_return


def _t_log_method_exception(_LogEntry):
    def _on_method_exception(ctx):
        logger.debug("Running arskom.web.base.on_method_exception() "
                                                                  "for logging")

        sc = ctx.service_class
        if sc is not None and sc.is_auxiliary():
            return

        ctx.event.call_end = time()

        logger.debug("logging method exception for %r" %
                                                      ctx.method_request_string)

        log_entry = _LogEntry()
        log_entry.ro = True

        try:
            out_data = ''.join(ctx.out_string)
        except:
            out_data = None
        else:
            # ctx.out_string generator ise az once kendisini tukettik. bari
            # tukettigimiz veriyi yerine koyalim ki bizden sonrakiler
            # kullanabilsin. degil idiyse de zarari yok bir suru stringi tek
            # string olarak birlestirdik.
            ctx.out_string = (out_data,)

        _fill_log(ctx, log_entry, 1, out_data)

    return _on_method_exception



#!/usr/bin/env python
# encoding: utf8

from __future__ import print_function

import os
import re
import sys

from setuptools import setup
from setuptools import find_packages
from setuptools.command.test import test as TestCommand


try:
    import colorama

    colorama.init()
    from colorama import Fore

    RESET = Fore.RESET
    GREEN = Fore.GREEN
    RED = Fore.RED
except ImportError:
    RESET = ''
    GREEN = ''
    RED = ''

import inspect
from os.path import join, dirname, abspath

OWN_PATH = abspath(inspect.getfile(inspect.currentframe()))
EXAMPLES_DIR = join(dirname(OWN_PATH), 'examples')

v = open(os.path.join(os.path.dirname(__file__), 'neurons', '__init__.py'), 'r')
VERSION = re.match(r".*__version__ = '(.*?)'", v.read(), re.S).group(1)

SHORT_DESC = \
    "Neurons is a web framework tying together Spyne, Twisted and SQLAlchemy."
LONG_DESC = SHORT_DESC

try:
    os.stat('CHANGELOG.rst')
    LONG_DESC += "\n\n" + open('CHANGELOG.rst', 'r').read()
except OSError:
    pass


common_reqs = ('spyne>=2.12', 'SQLAlchemy', 'Twisted>=15.2',
    'lxml>=3.8.0', 'pyyaml', 'msgpack-python', 'pycrypto', 'slimit',
    'txpostgres',
)


test_reqs = common_reqs + ('pytest', 'pytest-cov', 'pytest-twisted',
    'tox',
)

install_reqs = common_reqs + (
    'werkzeug', 'psycopg2>=2.5', 'txpostgres',
)

##################
# testing stuff

class Tox(TestCommand):
    user_options = [('tox-args=', 'a', "Arguments to pass to tox")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import tox
        import shlex

        errno = tox.cmdline(args=shlex.split(self.tox_args))
        sys.exit(errno)

# testing stuff
##################

# try the fix in:
# http://scipy-dev.scipy.narkive.com/pAiap0yr/scikits-contribution#post14
# if python setup.py sdist upload fails

setup(
    name='neurons',
    packages=find_packages(),

    version=VERSION,
    description=SHORT_DESC,
    long_description=LONG_DESC,
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    keywords=('http',),
    author='Burak Arslan',
    author_email='burak+neurons@arskom.com.tr',
    maintainer='Burak Arslan',
    maintainer_email='burak+neurons@arskom.com.tr',
    url='http://spyne.io/neurons',
    license='LGPL-2.1',
    zip_safe=False,
    install_requires=install_reqs,

    entry_points={
        'console_scripts': [],
    },

    package_data={
        'neurons.form.const': ['*.html'],
        'neurons.polymer.const': ['*.html', "*.js"],
        'neurons.polymer.const.comp': ['neurons-*/*'],
        'neurons.daemon.dowser.const': ['*.html', '*.css'],
    },

    tests_require=test_reqs,
    cmdclass={'test': Tox},
)



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import logging
logger = logging.getLogger(__name__)

from time import time

from contextlib import closing
from collections import namedtuple

from spyne import Integer32, M, Unicode
from neurons import TableModel


MigrationOperation = namedtuple("MigrationOperation",
                        "submodule migration_dict current_version migrate_init")


class Version(TableModel):
    migopts = []

    __tablename__ = "neurons_version"
    _type_info = [
        ('id', Integer32(pk=True)),
        ('submodule', M(Unicode, unique=True)),
        ('version', M(Integer32)),
    ]

    @classmethod
    def register_submodule(cls, submodule, migration_dict, current_version,
                                                             migrate_init=None):
        cls.migopts.append(
            MigrationOperation(submodule, migration_dict, current_version,
                                                                   migrate_init)
        )

    @staticmethod
    def migrate_all(config):
        for migopt in Version.migopts:
            Version.migrate(config, *migopt)

    @staticmethod
    def migrate(config, submodule, migration_dict, current_version,
                                                                  migrate_init):
        num_migops = 0
        table_name = Version.Attributes.table_name
        Version.Attributes.sqla_table.create(checkfirst=True)

        db = config.get_main_store()
        with closing(db.Session()) as session:
            logger.info("Acquiring pg_advisory_xact_lock(0) "
                                                    "for schema version checks")
            session.connection().execute("select pg_advisory_xact_lock(0)")

            # Create missing tables
            TableModel.Attributes.sqla_metadata.create_all(checkfirst=True)

            db_version = session.query(Version) \
                                         .filter_by(submodule=submodule).first()

            if db_version is None:
                version_entry = \
                           Version(submodule=submodule, version=current_version)

                session.add(version_entry)

                if migrate_init is not None:
                    migrate_init(config, session)
                    logger.info("Submodule '%s' schema version management "
                               "pre-init was executed successfully.", submodule)

                logger.info("Submodule '%s' schema version management "
                                        "was initialized as %d successfully.",
                                                     submodule, current_version)

                session.commit()

                return

            keys = [vernum for vernum in migration_dict.keys()
                              if db_version.version < vernum <= current_version]
            keys.sort()

            if len(keys) > 0:
                logger.info("%s schema version detected as %s. "
                                "Migration operation(s) %r will be performed",
                                            submodule, db_version.version, keys)

            for vernum in keys:
                migrate = migration_dict[vernum]

                with closing(db.Session()) as inner_session:
                    inner_db_version = inner_session.query(Version) \
                                           .filter_by(submodule=submodule).one()
                    inner_db_version.version = vernum

                    try:
                        start_t = time()
                        migrate(config, inner_session)

                    except Exception as e:
                        logger.exception(e)
                        logger.error("Migration operation %d failed, "
                                                     "stopping reactor", vernum)

                        from twisted.internet import reactor
                        from twisted.internet.task import deferLater
                        deferLater(reactor, 0, reactor.stop)
                        raise

                    inner_session.commit()

                num_migops += 1
                logger.info("%s schema migration to version %d took %.1fs.",
                                            submodule, vernum, time() - start_t)

        if num_migops == 0:
            logger.info("%s schema version detected as %s. Table unlocked.",
                                                     submodule, current_version)

        elif num_migops > 0:
            logger.info("%s schema version upgraded to %s "
                           "after %d migration operations. Table unlocked.",
                                         submodule, current_version, num_migops)



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import print_function

import logging
logger = logging.getLogger(__name__)

from inspect import isgenerator

from lxml.builder import E

from spyne import ComplexModelBase, Array
from spyne.util import Break, coroutine
from spyne.protocol.html import HtmlColumnTable

from neurons.form import HtmlForm
from neurons.form import HtmlFormRoot


SOME_COUNTER = [0]


# FIXME: Could NOTHING be done about the horrendous number of arguments here??!
class HtmlFormTable(HtmlColumnTable, HtmlFormRoot):
    def __init__(self, app=None, ignore_uncap=False, ignore_wrappers=True,
                     cloth=None, polymorphic=True, doctype=None, hier_delim='.',
                        cloth_parser=None, header=True, table_name_attr='class',
                        table_name=None, input_class=None, input_div_class=None,
                   input_wrapper_class=None, label_class=None, action=None,
                 table_class=None, field_name_attr='class', border=0,
                 row_class=None, cell_class=None, header_cell_class=None,
                 can_add=True, can_remove=True, label=True, before_table=None):

        super(HtmlFormTable, self).__init__(app=app,
                     ignore_uncap=ignore_uncap, ignore_wrappers=ignore_wrappers,
                polymorphic=polymorphic, hier_delim=hier_delim, doctype=doctype,
                          cloth=cloth, cloth_parser=cloth_parser, header=header,
                         table_name_attr=table_name_attr, table_name=table_name,
                       table_class=table_class, field_name_attr=field_name_attr,
                      border=border, row_class=row_class, cell_class=cell_class,
                 header_cell_class=header_cell_class, before_table=before_table)

        self.prot_form = HtmlForm(label=label, label_class=label_class,
                                        action=action, input_class=input_class,
                                                input_div_class=input_div_class,
                                        input_wrapper_class=input_wrapper_class)

        self.can_add = can_add
        self.can_remove = can_remove
        self.use_global_null_handler = False
        self.label = label
        self._init_input_vars(input_class, input_div_class,
                                               input_wrapper_class, label_class)

    @coroutine
    def model_base_to_parent(self, ctx, cls, inst, parent, name, from_arr=False,
                             remove=None, **kwargs):
        if from_arr:
            td_attrs = {}
            if self.field_name_attr:
                td_attrs[self.field_name_attr] = name

            with parent.element('tr'):
                with parent.element('td', attrib=td_attrs):
                    ret = self.prot_form.to_parent(ctx, cls, inst, parent, name,
                                     from_arr=from_arr, no_label=True, **kwargs)
                    if isgenerator(ret):
                        try:
                            while True:
                                y = (yield)
                                ret.send(y)

                        except Break as b:
                            try:
                                ret.throw(b)
                            except StopIteration:
                                pass

                if remove is None:
                    remove = True
                self.extend_data_row(ctx, cls, inst, parent, name,
                                                        remove=remove, **kwargs)

        else:
            ret = self.prot_form.to_parent(ctx, cls, inst, parent, name,
                                     from_arr=from_arr, no_label=True, **kwargs)

            if isgenerator(ret):
                try:
                    while True:
                        y = (yield)
                        ret.send(y)

                except Break as b:
                    try:
                        ret.throw(b)
                    except StopIteration:
                        pass

    # sole reason to override is to generate labels
    @coroutine
    def wrap_table(self, ctx, cls, inst, parent, name, gen_rows, **kwargs):
        if name == "":
            name = cls.get_type_name()

        cls_attr = self.get_cls_attrs(cls)
        if cls_attr.write is False:
            ctx.protocol.can_add = False
            ctx.protocol.can_remove = False
        else:
            ctx.protocol.can_add = self.can_add
            ctx.protocol.can_remove = self.can_remove

        if self.label:
            attrib = self._gen_label_wrapper_class(ctx, cls, name)
            with parent.element('div', attrib):
                parent.write(E.label(self.trc(cls, ctx.locale, name)))
                ret = self._gen_table(ctx, cls, inst, parent, name, gen_rows,
                                                                       **kwargs)
        else:
            ret = self._gen_table(ctx, cls, inst, parent, name, gen_rows,
                                                                       **kwargs)

        if isgenerator(ret):
            try:
                while True:
                    sv2 = (yield)
                    ret.send(sv2)
            except Break as b:
                try:
                    ret.throw(b)
                except StopIteration:
                    pass

    def extend_header_row(self, ctx, cls, parent, name, **kwargs):
        if ctx.protocol.can_add or self.can_remove:
            parent.write(E.th(**{'class': 'array-button'}))

    def extend_data_row(self, ctx, cls, inst, parent, name, array_index=None,
                                             add=False, remove=True, **kwargs):
        if array_index is None:
            return

        if not (ctx.protocol.can_remove or ctx.protocol.can_add):
            return

        td = E.td(style='white-space: nowrap;')
        td.attrib['class'] = 'array-button'
        self._gen_buttons(td, name, add, remove)

        parent.write(td)

    def _gen_buttons(self, elt, name, add, remove):
        name = "%s-%d" % (self.selsafe(name), SOME_COUNTER[0])

        if add:
            elt.append(
                E.button('+', **{
                    "class": "%s_btn_add" % name,
                    "type": "button",
                    "style": 'min-width:2em',
                }),
            )

        if remove:
            elt.append(
                E.button('-', **{
                    "class": "%s_btn_remove" % name,
                    "type": "button",
                    "style": 'min-width:2em',
                }),
            )

    def extend_table(self, ctx, cls, parent, name, **kwargs):
        if not ctx.protocol.can_add:
            return

        # FIXME: just fix me.
        if issubclass(cls, Array):
            cls = next(iter(cls._type_info.values()))
            self._gen_row(ctx, cls, [(cls.__orig__ or cls)()], parent, name)

        # even though there are calls to coroutines here, as the passed objects
        # are empty, so it's not possible to have push data. so there's no need
        # to check the returned generators.

        from_arr = True
        if issubclass(cls, ComplexModelBase):
            inst = (cls.__orig__ or cls)()
            ctx.protocol.inst_stack.append((cls, inst, from_arr))

            if cls.Attributes.max_occurs > 1:
                self._gen_row(ctx, cls, [inst], parent, name,
                                         add=True, remove=False, array_index=-1)
            else:
                self._gen_row(ctx, cls, inst, parent, name,
                                         add=True, remove=False, array_index=-1)

        else:
            inst = None
            ctx.protocol.inst_stack.append((cls, inst, from_arr))

            if cls.Attributes.max_occurs > 1:
                self.model_base_to_parent(ctx, cls, inst, parent, name,
                          from_arr=True, add=True, remove=False, array_index=-1)
            else:
                self.model_base_to_parent(ctx, cls, inst, parent, name,
                                         add=True, remove=False, array_index=-1)

        ctx.protocol.inst_stack.pop()

        name = '%s-%d' % (self.selsafe(name), SOME_COUNTER[0])

        SOME_COUNTER[0] += 1

        parent.write(self._format_js(ADD_JS, name=name))


ADD_JS = ["""
var add = function(event) {
    var td = $(this).parent();
    var tr = td.parent();
    var f = orig_row.clone(true);
    var btn = f.find(".%(name)s_btn_add");
    var pa = tr.parent();

    btn.off("click");
    btn.click(remove);
    btn.text("-");

    f.insertBefore(tr);
    rearrange();

    if (pa.find("tr:visible").length == 2) {
        var form = pa.closest('form');
        form.find('#' + root_name + '-empty').remove();
    }

    event.preventDefault();
    return false;
};

var remove = function (e) {
    var td = $(this).parent();
    var tr = td.parent();
    var pa = tr.parent();

    tr.remove();
    rearrange();

    if (pa.find("tr:visible").length == 1) {
        var form = pa.closest('form');
        $('<input>').attr({'type':'hidden', 'name': root_name, 'value': 'empty',
                                    'id': root_name + '-empty'}).appendTo(form);
    }

    e.preventDefault();
    return false;
};

var rearrange = function() {
    function zerofill(num, numZeros) {
        var n = Math.abs(num);
        var zeros = Math.max(0, numZeros - Math.floor(n).toString().length);
        var zeroString = Math.pow(10, zeros).toString().substr(1);
        if (num < 0) {
            zeroString = '-' + zeroString;
        }

        return zeroString + n;
    }

    table.children('tr').each(function() {
        var idx = $(this).index();
        $(this).find('[name]').each(function() {
            var elt = $(this);
            var name_str = elt.attr('name').replace("[-1]",
                                                  "[" + zerofill(idx, 9) + "]");
            elt.attr('name', name_str);
            console.log(elt + " name set to " + name_str);
        });
    });
};

$(".%(name)s_btn_remove").click(remove);
var btn_add = $(".%(name)s_btn_add");
btn_add.off("click");
btn_add.click(add);

var row = btn_add.parent().parent();
var table = row.parent();
var orig_row = row.clone(true);
var root_name = $(orig_row.find(":input")[0]).attr('name');
root_name = root_name.replace(new RegExp("\\\\[-1\\\\]\\\\..*"), "");

for (var children = row.children(), i=0, l=children.length-1; i < l; ++i) {
    $(children[i]).children().remove();
}

row.attr('null-widget', true);
var null_widget = row.clone(true);
"""]



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

import re

from spyne import Service, rpc, Unicode, ComplexModel, XmlAttribute
from spyne.protocol.cloth import XmlCloth

from slimit.mangler import mangle as slimit_mangler
from slimit.visitors.minvisitor import ECMAMinifier

from neurons.base.screen import Link
from neurons.polymer.jsutil import get_js_parser, set_js_variable
from neurons.polymer.const import POLYMER_PREAMBLE, DEFAULT_URL_POLYFILL
from neurons.polymer.const import POLYMER_DEFN_TEMPLATE

from .model import PolymerComponent, HtmlImport, IronAjax, PolymerScreen


def _to_snake_case(name, delim='-'):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1%s\2' % delim, name)
    return re.sub('([a-z0-9])([A-Z])', r'\1%s\2' % delim, s1).lower()


def gen_component_imports(deps):
    for comp in deps:
        if isinstance(comp, HtmlImport):
            yield comp
            continue

        if '/' in comp:
            yield HtmlImport(
                href="/static/bower_components/{0}.html".format(comp)
            )

        else:
            yield HtmlImport(
                href="/static/bower_components/{0}/{0}.html".format(comp)
            )


def gen_polymer_defn(component_name, cls, minify=False):
    parser = get_js_parser()
    tree = parser.parse(POLYMER_DEFN_TEMPLATE)

    entries = tree.children()[0].children()[0].children()[1].children()

    # set tag name
    e0 = entries[0]
    assert e0.left.value == 'is'
    e0.right.value = '"{}"'.format(component_name)

    if minify:
        slimit_mangler(tree, toplevel=True)
        retval = ECMAMinifier().visit(tree)
    else:
        retval = tree.to_ecma()

    return retval


def gen_component(cls, component_name, DetailScreen, gen_css_imports, minify_js,
                                     api_read_url_prefix, api_write_url_prefix):
    deps = [
        'polymer',

        'iron-ajax',
        'iron-form',

        'paper-input',
        'paper-input/paper-textarea',
        'paper-checkbox',

        # required for dropdown menu
        'paper-item',
        'paper-listbox',

        'paper-dropdown-menu',
        'paper-dropdown-menu/paper-dropdown-menu-light',

        HtmlImport(href='neurons-array.html'),
        HtmlImport(href='neurons-date-time/neurons-datetime-picker.html'),
        HtmlImport(href='neurons-complex-reference/neurons-complex-dropdown.html'),
        HtmlImport(href='neurons-complex-reference/neurons-complex-href.html'),
    ]

    styles = []
    if gen_css_imports:
        styles.append('@import url("/static/screen/{}.css")'
                                                        .format(component_name))

    getter_url = "{}{}.get".format(api_read_url_prefix, cls.get_type_name())
    putter_url = "{}{}.put".format(api_write_url_prefix, cls.get_type_name())

    retval = DetailScreen(
        main=(cls.__orig__ or cls)(),
        getter=IronAjax(url=getter_url),
        putter=IronAjax(url=putter_url),
        dom_module_id=component_name,
        definition=gen_polymer_defn(component_name, cls, minify=minify_js),
        dependencies=gen_component_imports(deps),
    )

    if len(styles) > 0:
        retval.style = '\n'.join(styles)

    return retval


def TComponentGeneratorService(cls, prefix=None, locale=None,
                     gen_css_imports=False, minify_js=False,
                               api_read_url_prefix='', api_write_url_prefix=''):
    type_name = cls.get_type_name()
    component_name = _to_snake_case(type_name)

    if prefix is not None:
        component_name = "{}-{}".format(prefix, component_name)

    method_name = component_name + ".html"

    class DetailScreen(PolymerComponent):
        __type_name__ = '{}DetailComponent'.format(type_name)
        main = cls

    class ComponentGeneratorService(Service):
        @rpc(Unicode(6), _returns=DetailScreen, _body_style='out_bare',
            _in_message_name=method_name,
            _internal_key_suffix='_' + component_name)
        def _gen_component(ctx, locale):
            if locale is not None:
                ctx.locale = locale
                logger.debug("Locale overridden to %s locally.", locale)
            return gen_component(cls, component_name, DetailScreen,
                             gen_css_imports, minify_js, api_read_url_prefix,
                                                         api_write_url_prefix)

    if locale is not None:
        def _fix_locale(ctx):
            if ctx.locale is None:
                ctx.locale = 'tr_TR'
                logger.debug("Locale overridden to %s.", ctx.locale)

        ComponentGeneratorService.event_manager \
                                       .add_listener('method_call', _fix_locale)

    return ComponentGeneratorService


def _get_getter_input(cls):
    try:
        getter_descriptor = cls.Attributes.methods['get']
    except KeyError:
        raise Exception("%r needs a getter", cls)

    return getter_descriptor.in_message


class ScreenParams(ComplexModel):
    locale = Unicode(6)


def TScreenGeneratorService(cls, prefix=None, url_polyfill=DEFAULT_URL_POLYFILL,
              url_service_worker=None, minify_js=False, comp_prefix=''):
    type_name = cls.get_type_name()
    component_name = _to_snake_case(type_name)

    if prefix is not None:
        component_name = "{}-{}".format(prefix, component_name)

    method_name = component_name + ".html"

    getter_in_cls = _get_getter_input(cls)

    class DetailScreen(PolymerScreen):
        main = getter_in_cls.customize(
            protocol=XmlCloth(),
            sub_name=component_name,
        )

    class ScreenGeneratorService(Service):
        @rpc(getter_in_cls, ScreenParams, _returns=PolymerScreen,
            _body_style='out_bare', _in_message_name=method_name,
            _internal_key_suffix='_' + component_name)
        def _gen_screen(ctx, query, params):
            if query is None:
                query = getter_in_cls()

            retval = DetailScreen(ctx, title=type_name, main=query) \
                .with_jquery() \
                .with_xml_to_jsobj() \
                .with_type_checkers() \
                .with_clone() \
                .with_urlencode()

            if params is not None and params.locale is not None:
                retval.links = [
                    Link(
                        rel='import',
                        href="{}/{}?locale={}".format(comp_prefix, method_name,
                                                                  params.locale)
                    )
                ]

            else:
                retval.links = [
                    Link(
                        rel='import',
                        href="{}/{}".format(comp_prefix, method_name)
                    ),
                ]

            data = {'url_polyfill': url_polyfill}
            if url_service_worker is not None:
                data['service_worker_url'] = url_service_worker

            tree = get_js_parser().parse(POLYMER_PREAMBLE)
            set_js_variable(tree, 'polymer_init_options', data)
            retval.append_script_tree(tree, minify=minify_js)

            return retval

    return ScreenGeneratorService



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


from __future__ import print_function, absolute_import

import logging
logger = logging.getLogger(__name__)

import os
import re

from threading import Lock
from os.path import abspath, join, isfile

from colorama import Fore, Style

from spyne import Application, UnsignedInteger, ComplexModel, Unicode, \
    UnsignedInteger16, Boolean, String, Array, ComplexModelBase, M, \
    ValidationError, Integer32
from spyne.util.resource import get_resource_path

from neurons.daemon import EXIT_ERR_LISTEN_TCP, EXIT_ERR_LISTEN_UDP, \
                                                                EXIT_ERR_UNKNOWN
from neurons.daemon.cli import config_overrides
from neurons.daemon.config._wdict import wdict, wrdict, Twrdict


class Service(ComplexModel):
    name = M(Unicode)
    disabled = Boolean(default=False)

    def __init__(self, *args, **kwargs):
        self._parent = None
        self.color = None

        super(Service, self).__init__(*args, **kwargs)

    def set_parent(self, parent):
        assert self._parent is None
        assert parent is not None

        self._parent = parent

    def _parse_overrides(self):
        pass

    @property
    def colored_name(self):
        if self.color is None:
            return self.name

        return '%s%s[%s]%s' % (self.color, Style.BRIGHT, self.name,
                                                                Style.RESET_ALL)


# noinspection PyPep8Naming
def _TFactoryProxy():
    from twisted.internet.protocol import ServerFactory

    class FactoryProxy(ServerFactory):
        def __init__(self):
            # real_factory is set by neurons.daemon.main
            self.real_factory = None
            self.run_start_factory = False
            self.noisy = False

        @classmethod
        def forProtocol(cls, *args, **kwargs):
            raise NotImplementedError()

        @property
        def real_factory(self):
            return self.__rf

        @real_factory.setter
        def real_factory(self, what):
            self.__rf = what
            if self.run_start_factory:
                what.startFactory()

        def logPrefix(self):
            if self.real_factory is not None:
                return self.real_factory.logPrefix()
            return "EmptyFactoryProxy"

        def startFactory(self):
            if self.real_factory is not None:
                return self.real_factory.startFactory()
            self.run_start_factory = True

        def stopFactory(self):
            if self.real_factory is not None:
                return self.real_factory.stopFactory()

        def buildProtocol(self, addr):
            return self.real_factory.buildProtocol(addr)

    return FactoryProxy


_lock_factory_proxy = Lock()


class Client(Service):
    host = Unicode(nullable=False, default='0.0.0.0')
    port = UnsignedInteger16(nullable=False, default=0)
    type = M(Unicode(values=('tcp4', 'tcp6', 'udp4', 'udp6', 'unix'),
                                                                default='tcp4'))

    def _parse_overrides(self):
        super(Client, self)._parse_overrides()

        for k, v in config_overrides.items():
            if k.startswith('--host-%s' % self.name):
                host = v
                self.host = host
                logger.debug("Overriding host for server service '%s' to '%s'",
                                                           self.name, self.host)

                del config_overrides[k]
                continue

            if k.startswith('--port-%s' % self.name):
                port = v
                self.port = int(port)
                logger.debug("Overriding port for server service '%s' to '%s'",
                                                           self.name, self.port)
                del config_overrides[k]
                continue



class Server(Service):
    backlog = Integer32(default=50)

    host = Unicode(nullable=False, default='0.0.0.0')
    port = UnsignedInteger16(nullable=False, default=0)
    type = M(Unicode(values=('tcp4', 'tcp6', 'udp4', 'udp6', 'unix'),
                                                                default='tcp4'))

    FactoryProxy = None

    def __init__(self, *args, **kwargs):
        super(Server, self).__init__(*args, **kwargs)

        self.d = None
        self.listener = None
        self.failed = False
        self.color = Fore.YELLOW  # set to G by daemon.main._set_real_factory

    def gen_endpoint(self, reactor):
        # FIXME: We might not need endpoints after all..
        if self.type == 'tcp4':
            from twisted.internet.endpoints import TCP4ServerEndpoint
            return TCP4ServerEndpoint(reactor, port=self.port,
                                      backlog=self.backlog, interface=self.host)

        elif self.type == 'tcp6':
            from twisted.internet.endpoints import TCP6ServerEndpoint
            return TCP6ServerEndpoint(reactor, port=self.port,
                                      backlog=self.backlog, interface=self.host)
        elif self.type == 'udp4':
            # TODO
            raise NotImplementedError(self.type)
            #return UDP4ServerEndpoint(reactor, port=self.port,
            #                         backlog=self.backlog, interface=self.host)
        elif self.type == 'udp6':
            # TODO
            raise NotImplementedError(self.type)
            #return UDP6ServerEndpoint(reactor, port=self.port,
            #                         backlog=self.backlog, interface=self.host)

        raise ValidationError(self.type)

    def get_factory_proxy(self):
        # Why thread-safe application of daemon configuration? I say why not :)
        with _lock_factory_proxy:
            if Server.FactoryProxy is None:
                Server.FactoryProxy = _TFactoryProxy()

        return Server.FactoryProxy

    @property
    def lstr(self):
        return "{}:{}:{}".format(self.type.upper(), self.host, self.port)

    def listen(self):
        from twisted.internet import reactor

        FactoryProxy = self.get_factory_proxy()

        retval = self.d = self.gen_endpoint(reactor) \
            .listen(FactoryProxy()) \
                .addCallback(self.set_listening_port) \
                .addCallback(lambda _: logger.info("%s listening on %s",
                                                self.colored_name, self.lstr)) \
                .addErrback(self._eb_listen)

        return retval

    def _eb_listen(self, err):
        self.failed = True

        self.color = Fore.RED
        logging.error("%s Error listening to %s, stopping reactor\n%s",
                               self.colored_name, self.lstr, err.getTraceback())

        if self.type.startswith('udp'):
            raise os._exit(EXIT_ERR_LISTEN_UDP + self.port)

        elif self.type.startswith('tcp'):
            raise os._exit(EXIT_ERR_LISTEN_TCP + self.port)

        raise os._exit(EXIT_ERR_UNKNOWN)

    def set_listening_port(self, listening_port):
        assert not (listening_port is None)
        self.listener = listening_port

    def _parse_overrides(self):
        super(Server, self)._parse_overrides()

        for k, v in config_overrides.items():
            if k.startswith('--host-%s' % self.name):
                host = v
                self.host = host
                logger.debug("Overriding host for server service '%s' to '%s'",
                                                           self.name, self.host)

                del config_overrides[k]
                continue

            if k.startswith('--port-%s' % self.name):
                port = v
                self.port = int(port)
                logger.debug("Overriding port for server service '%s' to '%s'",
                                                           self.name, self.port)
                del config_overrides[k]
                continue


def _listfiles(dirpath):
    for f in os.listdir(dirpath):
        fp = join(dirpath, f)
        if isfile(fp):
            yield fp


class SslServer(Server):
    _type_info = [
        ('cacert_path', Unicode),
        ('cacert', Unicode),
        ('cert', Unicode),
        ('key', Unicode),
        ('verify', Boolean(default=False)),
        ('verdepth', Integer32(default=9)),  # taken from twisted default
    ]

    @staticmethod
    def get_path(s):
        if s.startswith("rsc:"):
            package, file_name = s[4:].split("/", 1)
            return get_resource_path(package, file_name)
        return s

    def gen_endpoint(self, reactor):
        from OpenSSL import crypto

        cert = None
        if self.cert is not None:
            fn = self.get_path(self.cert)
            logger.debug("%s Loading cert from: %s", self.colored_name, fn)
            fdata = open(fn, 'rb').read()
            cert = crypto.load_certificate(crypto.FILETYPE_PEM, fdata)
        else:
            logger.warning("%s No cert loaded", self.colored_name)


        key = None
        if self.key is not None:
            fn = self.get_path(self.key)
            logger.debug("%s Loading key from: %s", self.colored_name, fn)
            fdata = open(fn, 'rb').read()
            key = crypto.load_privatekey(crypto.FILETYPE_PEM, fdata)
        else:
            logger.warning("%s No key loaded", self.colored_name)

        cacerts = []
        if self.cacert is not None:
            fn = self.get_path(self.key)
            logger.debug("%s Loading cacert from: %s", self.colored_name, fn)
            fdata = open(fn, 'rb').read()
            cacert = crypto.load_certificate(crypto.FILETYPE_PEM, fdata)
            cacerts.append(cacert)

        if self.cacert_path is not None:
            cacert_path = self.get_path(self.key)
            # TODO: ignore errors
            for fn in _listfiles(cacert_path):
                fdata = open(fn, 'rb').read()
                logger.debug("%s Loading cacert from: %s",
                                                          self.colored_name, fn)
                cacert = crypto.load_certificate(crypto.FILETYPE_PEM, fdata)
                cacerts.append(cacert)

        from twisted.internet.ssl import CertificateOptions
        options = CertificateOptions(
            privateKey=key,
            certificate=cert,
            caCerts=cacerts,
            verify=self.verify,
            verifyDepth=self.verdepth,
        )

        assert options.getContext()

        if self.type == 'tcp4':
            from twisted.internet.endpoints import SSL4ServerEndpoint
            return SSL4ServerEndpoint(reactor, self.port, options,
                                      backlog=self.backlog, interface=self.host)

        elif self.type == 'tcp6':
            from twisted.internet.endpoints import SSL6ServerEndpoint
            return SSL6ServerEndpoint(reactor, self.port, options,
                                      backlog=self.backlog, interface=self.host)

        raise ValidationError(self.type)


class HttpApplication(ComplexModel):
    url = Unicode

    def __init__(self, app=None, url=None):
        super(HttpApplication, self).__init__(url=url)
        self.app = app

    def gen_resource(self):
        from spyne.server.twisted import TwistedWebResource
        from spyne.server.wsgi import WsgiApplication
        from spyne.util.wsgi_wrapper import WsgiMounter

        from twisted.internet import reactor
        from twisted.web.resource import Resource
        from twisted.web.wsgi import WSGIResource

        if isinstance(self.app, Resource):
            retval = self.app

        elif isinstance(self.app, Application):
            retval = TwistedWebResource(self.app)

        elif isinstance(self.app, (WsgiApplication, WsgiMounter)):
            retval = WSGIResource(reactor, reactor.getThreadPool(), self.app)

        else:
            raise ValueError(self.app)

        retval.prepath = self.url
        return retval


class StaticFileServer(HttpApplication):
    path = String
    list_contents = Boolean(default=False)
    disallowed_exts = Array(Unicode, default_factory=tuple)

    def __init__(self, *args, **kwargs):
        # We need the default ComplexModelBase ctor and not HttpApplication's
        # custom ctor here

        ComplexModelBase.__init__(self, *args, **kwargs)

    def gen_resource(self):
        if self.list_contents:
            from neurons.daemon.config.static_file import TCheckedFile
            CheckedFile = TCheckedFile(self.disallowed_exts, self.url)

            return CheckedFile(abspath(self.path))

        from neurons.daemon.config.static_file import TStaticFile
        StaticFile = TStaticFile(self.disallowed_exts, self.url)

        return StaticFile(abspath(self.path))


class HttpServer(Server):
    _type_info = [
        ('static_dir', Unicode),
        ('_subapps', Array(HttpApplication, sub_name='subapps')),
    ]

    def _push_asset_dir_overrides(self, obj):
        if obj.url == '':
            key = '--assets-%s=' % self.name
        else:
            suburl = re.sub(r'[\.-/]+', obj.url, '-')
            key = '--assets-%s-%s=' % (self.name, suburl)

        for a in config_overrides:
            if a.startswith(key):
                _, dest = a.split('=', 1)
                obj.path = abspath(dest)

                logger.debug("Overriding asset path for app "
                         "'%s', url '%s' to '%s'", self.name, obj.url, obj.path)

    def __init__(self, *args, **kwargs):
        super(HttpServer, self).__init__(*args, **kwargs)

    def gen_site(self):
        from twisted.web.server import Site

        subapps = []
        for url, subapp in self.subapps.items():
            if isinstance(subapp, StaticFileServer):
                self._push_asset_dir_overrides(subapp)

            if isinstance(subapp, HttpApplication):
                assert subapp.url is not None, subapp.app
                if hasattr(subapp, 'app'):
                    if subapp.app is None:
                        logger.warning("No subapp '%s' found in app '%s': "
                                       "Invalid key.", subapp.url, self.name)
                        continue
                subapps.append(subapp)

            else:
                subapps.append(HttpApplication(subapp, url=url))

        self._subapps = subapps

        root_app = self.subapps.get('', None)

        if root_app is None:
            from spyne.server.twisted.http import get_twisted_child_with_default
            from twisted.web.resource import Resource

            class TwistedResource(Resource):
                def getChildWithDefault(self, path, request):
                    return get_twisted_child_with_default(self, path, request)

            root = TwistedResource()
            root.prepath = '/'
        else:
            root = root_app.gen_resource()

        for subapp in self._subapps:
            if subapp.url != '':
                root.putChild(subapp.url, subapp.gen_resource())

        retval = Site(root)

        retval.displayTracebacks = self._parent.debug
        return retval

    @property
    def _subapps(self):
        if self.subapps is not None:
            for k, v in self.subapps.items():
                v.url = k

            return self.subapps.values()

        self.subapps = wrdict()
        return []

    @_subapps.setter
    def _subapps(self, subapps):
        if isinstance(subapps, dict):
            self.subapps = Twrdict(self, 'url')()
            for k, v in subapps.items():
                self.subapps[k] = v

        elif isinstance(subapps, (tuple, list)):
            self.subapps = Twrdict(self, 'url')()
            for v in subapps:
                assert v.url is not None, "%r.url is None" % v
                self.subapps[v.url] = v

        self.subapps = subapps
        if subapps is not None:
            self.subapps = wdict([(s.url, s) for s in subapps])


class WsgiServer(HttpServer):
    thread_min = UnsignedInteger
    thread_max = UnsignedInteger
